@@ -3,6 +3,7 @@
 ## ipsecr.fit.R
 ## 2022-04-01,18,19, 2022-05-08, 2022-06-11
 ## 2022-06-13 lambdak renamed NT
+## 2022-12-21 etc. allow extra parameters
 ###############################################################################
 
 ipsecr.fit <- function (
@@ -124,7 +125,9 @@ ipsecr.fit <- function (
         factorial    = 'full',
         FrF2args     = NULL,
         YonX         = TRUE,
-        param        = 0               ## needed by 'secr'
+        param        = 0,               ## needed by 'secr'
+        extraparam   = NULL,
+        forkonunix   = TRUE 
     )
     if (any(!(names(details) %in% names(defaultdetails)))) {
         stop ("details list includes invalid name")
@@ -163,8 +166,8 @@ ipsecr.fit <- function (
     if (ms(capthist)) {
         nsessions <- length(capthist)
         noccasions <- sapply(capthist, ncol)
-        if (details$popmethod == 'sim.pop' && packageVersion('secr') < '4.5.6') {
-            stop ("simulation of multi-session population with sim.pop requires secr >= 4.5.6")
+        if (details$popmethod == 'sim.popn' && packageVersion('secr') < '4.5.6') {
+            stop ("simulation of multi-session population with sim.popn requires secr >= 4.5.6")
         }
         if (!is.null(mask) && !ms(mask)) {
             ## inefficiently replicate mask for each session!
@@ -200,6 +203,8 @@ ipsecr.fit <- function (
         }
     }
     
+    if (details$debug) print(summary(capthist, terse = TRUE))
+    
     #################################################
     ## nontarget interference 
     #################################################
@@ -235,20 +240,26 @@ ipsecr.fit <- function (
     
     fnames <- names(fixed)
     
+    #################################################
+    ## parameter names
+    #################################################
+    pnames <- valid.pnames (details, FALSE, detectfn, FALSE, FALSE, 1)
+    extrapnames <- extraParNames(details, fixed)  # estimated user parameters
+    if (modelnontarget) pnames <- c(pnames, 'NT')
+
     #################################################
     ## build default model and update with user input
     #################################################
     
     defaultmodel <- list(D=~1, g0=~1, lambda0=~1, sigma=~1, z=~1, w=~1, NT=~1)
     defaultmodel <- replace (defaultmodel, names(model), model)
-    
-    #################################################
-    ## parameter names
-    #################################################
-    
-    pnames <- valid.pnames (details, FALSE, detectfn, FALSE, FALSE, 1)
-    if (modelnontarget) pnames <- c(pnames, 'NT')
-    
+    for (p in extrapnames) {
+        defaultmodel[[p]] <- ~1
+    }
+    for (f in fnames) {
+        if (f %in% names(details$extraparam)) details$extraparam[[f]] <- fixed[[f]]
+    }
+
     #################################################
     ## test for irrelevant parameters in user's model
     #################################################
@@ -276,7 +287,8 @@ ipsecr.fit <- function (
     #################################################
     
     defaultlink <- list(D = 'log', g0 = 'logit', lambda0 = 'log', sigma = 'log', NT = 'log')
-    link <- replace (defaultlink, names(link), link)
+    defaultextralink <- setNames(as.list(rep('log',length(extrapnames))), extrapnames)
+    link <- replace (c(defaultlink, defaultextralink), names(link), link)
     link[!(names(link) %in% pnames)] <- NULL
     
     ############################################
@@ -336,7 +348,14 @@ ipsecr.fit <- function (
     detbetanames <- detBetaNames(popn, model, detectfn, sessionlevels, fixed, details)
     nDetectionParameters <- sapply(detbetanames, length)
     
-    np <- c(D = nDensityParameters, nDetectionParameters, NT = nNTParameters)
+    if (length(extrapnames)>0) {
+        nExtraParameters <- sapply(extrapnames, length)
+    }
+    else {
+        nExtraParameters <- 0
+    }
+    
+    np <- c(D = nDensityParameters, nDetectionParameters, NT = nNTParameters, nExtraParameters)
     NP <- sum(np)
     parindx <- split(1:NP, rep(1:length(np), np))
     names(parindx) <- names(np)[np>0]
@@ -350,6 +369,10 @@ ipsecr.fit <- function (
     if (nNTParameters>0) {
         betanames <- c(betanames, paste('NT', NTnames, sep='.'))
     }
+    if (any(nExtraParameters>0)) {
+        betanames <- c(betanames, extrapnames)  # assume one beta per extrapname
+    }
+    
     betanames <- sub('..(Intercept))','',betanames)
     
     ###########################################
@@ -388,7 +411,15 @@ ipsecr.fit <- function (
             NT <- NULL
         }
         
-        repeat {
+        if (any(nExtraParameters>0)) {
+            for (parm in extrapnames) {
+                details$extraparam[[parm]] <- untransform (beta[parm], link[[parm]])
+            }
+        }
+       
+        # repeat {
+        allOK <- FALSE
+        while (attempts < details$max.ntries && !allOK) {
             
             #------------------------------------------------
             # generate population
@@ -402,91 +433,103 @@ ipsecr.fit <- function (
                 popn <- sim.popn (D = as.data.frame(D), core = mask, 
                     model2D = 'IHP', Ndist = Ndist, nsessions = nsessions)
             }
+            
+            if (details$debug) print(summary(popn))
+            
+            
             # abort if no animals to sample
             if ((nsessions == 1 && length(popn)<2) || (nsessions>1 && any(sapply(popn,length)<2))) {
                 warning ("ipsecr.fit: no animals in simulated population", call. = FALSE)
-                return(rep(NA, NP))
+                # return(rep(NA, NP))
             }
-            #-----------------------------------------------------
-            # session-specific detection parameter matrix/matrices
-            scalepopn <- function (popn, mask) {
-                if (ms(popn)) {
-                    out <- mapply(scalepopn, popn, mask, SIMPLIFY = FALSE)
-                    class(out) <- class(popn)
-                    out
+            else {
+                #-----------------------------------------------------
+                # session-specific detection parameter matrix/matrices
+                scalepopn <- function (popn, mask) {
+                    if (ms(popn)) {
+                        out <- mapply(scalepopn, popn, mask, SIMPLIFY = FALSE)
+                        class(out) <- class(popn)
+                        out
+                    } 
+                    else {
+                        meanSD <- attr(mask,'meanSD')
+                        popn[,] <- scale(popn, meanSD[1,], meanSD[2,])
+                        popn
+                    }
+                }
+                popn2 <- scalepopn(popn, mask)
+                detparmat <- getDetParMat (popn2, model, detectfn, beta, parindx, 
+                    link, fixed, details, sessionlevels)
+                #-----------------------------------------------------
+                if (details$debug) {
+                    if (ms(capthist)) 
+                        lapply(detparmat, function(x) print(apply(x,2,mean)))
+                    else 
+                        print(apply(detparmat,2,mean))
+                }
+                # sample from population
+                if (is.function(details$CHmethod)) {   # user
+                    ch <- details$CHmethod(trps, popn, detectfn, detparmat, noccasions, NT, details)
+                }
+                else if (details$CHmethod == 'internal') {   # C++
+                    ch <- simCH(trps, popn, detectfn, detparmat, noccasions, NT, details)
+                }
+                else if (details$CHmethod == 'sim.capthist') {   
+                    if (modelnontarget) stop ("sim.capthist does not simulate nontarget detections")
+                    ch <- sim.capthist(trps, popn, detectfn, as.list(detparmat[1,]), noccasions, nsessions, 
+                        renumber = FALSE)
+                }
+                # check valid
+                n <- if (nsessions == 1) nrow(ch) else sapply(ch,nrow)
+                r <- if (nsessions == 1) sum(abs(ch)>0) else sapply(ch, function(x) sum(abs(x)>0))
+                if (any(n==0)) {
+                    warning ("ipsecr.fit: no captures in simulation", call. = FALSE)
                 } 
                 else {
-                    meanSD <- attr(mask,'meanSD')
-                    popn[,] <- scale(popn, meanSD[1,], meanSD[2,])
-                    popn
+                    if (any((r - n) < 1))
+                        warning ("ipsecr.fit: no re-captures in simulation", call. = FALSE)
                 }
-            }
-            popn2 <- scalepopn(popn, mask)
-            detparmat <- getDetParMat (popn2, model, detectfn, beta, parindx, link, fixed, details, sessionlevels)
-            #-----------------------------------------------------
-
-            # sample from population
-            if (is.function(details$CHmethod)) {   # user
-                ch <- details$CHmethod(trps, popn, detectfn, detparmat, noccasions, NT, details)
-            }
-            else if (details$CHmethod == 'internal') {   # C++
-                ch <- simCH(trps, popn, detectfn, detparmat, noccasions, NT, details)
-            }
-            else if (details$CHmethod == 'sim.capthist') {   
-                if (modelnontarget) stop ("sim.capthist does not simulate nontarget detections")
-                ch <- sim.capthist(trps, popn, detectfn, as.list(detparmat[1,]), noccasions, nsessions, 
-                    renumber = FALSE)
-            }
-            # check valid
-            n <- if (nsessions == 1) nrow(ch) else sapply(ch,nrow)
-            r <- if (nsessions == 1) sum(abs(ch)>0) else sapply(ch, function(x) sum(abs(x)>0))
-            if (any(n==0)) {
-                warning ("ipsecr.fit: no captures in simulation", call. = FALSE)
-            } 
-            else {
-                if (any((r - n) < 1))
-                    warning ("ipsecr.fit: no re-captures in simulation", call. = FALSE)
-            }
-
-            #------------------------------------------------
-            # predict values
-            predicted <- try (proxyfn (ch, model = model, trapdesigndata = trapdesigndata, ...))
-            if (inherits(predicted, 'try-error')) {
-                predicted <- rep(NA, NP)
-            }
-            if (details$debug) {
-                # conditional on specified design point 
-                # (fails with ncores>1)
-                debug <- as.numeric(details$debug)
-                if (abs(debug)==1 || 
-                        (abs(debug)>1 && 
-                                isTRUE(all.equal(designbeta[abs(debug),], 
-                                    beta, tolerance = 1e-4)))) 
-                {
-                    cat('debug point ', abs(debug), '\n')
-                    cat('designbeta ', unlist(designbeta[abs(debug),]), '\n')
-                    cat('N ', N, '\n')
-                    cat('detparmat', unlist(detparmat), '\n')
-                    cat('detectfn', detectfn, '\n')
-                    print(summary(popn))
-                    print(summary(ch))
-                    cat('\npredicted', predicted, '\n')   # proxy vector
-                    
-                    # cat ('saving ch to ch.RDS\n')
-                    # saveRDS(ch, file='ch.RDS')
-                    # cat('\nproxyfn\n')
-                    # print(proxyfn)
-                    
-                    if (debug<0) stop()
+                if (details$debug) print(summary(ch))
+                
+                #------------------------------------------------
+                # predict values
+                predicted <- try (proxyfn (ch, model = model, trapdesigndata = trapdesigndata, ...))
+                if (inherits(predicted, 'try-error')) {
+                    predicted <- rep(NA, NP)
                 }
+                if (details$debug) {
+                    # conditional on specified design point 
+                    # (fails with ncores>1)
+                    debug <- as.numeric(details$debug)
+                    if (abs(debug)==1 || 
+                            (abs(debug)>1 && 
+                                    isTRUE(all.equal(designbeta[abs(debug),], 
+                                        beta, tolerance = 1e-4)))) 
+                    {
+                        cat('debug point ', abs(debug), '\n')
+                        cat('designbeta ', unlist(designbeta[abs(debug),]), '\n')
+                        cat('N ', N, '\n')
+                        cat('detparmat', unlist(detparmat), '\n')
+                        cat('detectfn', detectfn, '\n')
+                        print(summary(popn))
+                        print(summary(ch))
+                        cat('\npredicted', predicted, '\n')   # proxy vector
+                        
+                        # cat ('saving ch to ch.RDS\n')
+                        # saveRDS(ch, file='ch.RDS')
+                        # cat('\nproxyfn\n')
+                        # print(proxyfn)
+                        
+                        if (debug<0) stop()
+                    }
+                }
+                attempts <- attempts+1
+                
+                #------------------------------------------------
+                ## exit loop if exceeded max.ntries or all OK
+                allOK <- !any(is.na(predicted)) && all(is.finite(predicted))
+                # if (attempts >= details$max.ntries || allOK) break
             }
-            attempts <- attempts+1
-            
-            #------------------------------------------------
-            ## exit loop if exceeded max.ntries or all OK
-            allOK <- !any(is.na(predicted)) && all(is.finite(predicted))
-            if (attempts >= details$max.ntries || allOK)
-                break
         }
         #----------------------------------------------------
 
@@ -514,22 +557,25 @@ ipsecr.fit <- function (
     ## target values of predictor
     ##########################################
     y <- proxyfn(capthist, model = model, trapdesigndata = trapdesigndata, ...)
-    if (length(y) != NP)
-        stop ("need one proxy for each coefficient ",
+    # if (length(y) != NP)
+    if (length(y) < NP)
+        stop ("need at least one proxy for each coefficient ",
             paste(betanames, collapse=" "))
-
+    
     ##########################################
     ## starting values
     ##########################################
-    ## ad hoc exclusion of NT 2022-05-06
+    ## ad hoc exclusion of NT and extra parameters
     details$trace <- verbose  # as used by makeStart
-    start <- makeStart(start, parindx[names(parindx) != 'NT'], 
-        capthist, mask, detectfn, link, details, fixed)
+    start <- makeStart(start, parindx[!(names(parindx) %in% c('NT', extrapnames))], 
+        capthist, mask, detectfn, link, details[names(details) != 'extraparam'], fixed)
     if (modelnontarget) {
         start <- c(start, y[parindx$NT])
     }
-    
-    ############################################
+    if (length(extrapnames)>0) {
+        extrastart <- mapply(transform, details$extraparam[extrapnames], link[extrapnames], SIMPLIFY = FALSE)
+        start <- c(start, unlist(extrastart))
+    }
     # Fixed beta parameters
     ############################################
     fb <- details$fixedbeta
@@ -548,32 +594,29 @@ ipsecr.fit <- function (
     ###########################################
     
     if (ncores > 1) {
-        memo ('Preparing cluster for parallel processing', verbose)
-        if(.Platform$OS.type == "unix") {
-            # clust <- makeCluster(ncores, type = "FORK", outfile = "clusterlogfile.txt")
-            clust <- makeCluster(ncores, type = "FORK")
-        } 
-        else {
-            # clust <- makeCluster(ncores, type = "PSOCK", outfile = "clusterlogfile.txt")
-            clust <- makeCluster(ncores, type = "PSOCK")
+        clustertype <- if (details$forkonunix && .Platform$OS.type == "unix") "FORK" else "PSOCK"
+        clusterfile <- if (details$debug) "clusterlogfile.txt" else "/dev/null"
+        memo (paste0('Preparing ', clustertype, 
+            ' cluster for parallel processing (ncores = ', ncores, ')'), verbose)
+        clust <- makeCluster(ncores, type = clustertype, outfile = clusterfile,
+            methods = FALSE, useXDR = .Platform$endian=='big')
+        if (clustertype == "PSOCK") {
             clusterExport(clust, c(
                 "mask", "trps", "link", "fixed", "details", 
                 "detectfn", "noccasions", "nsessions", "proxyfn",
                 "model", "trapdesigndata", "parindx", 
                 "designD", "designNT", "modelnontarget", "cellarea"
-                # following are exported only during testing
-                # ,"getD", "untransform", "simpop", "simCH", "ms", "getDetParMat", 
-                # "parnames", "getDetDesignData", "covariates", "invlogit", 
-                # "usage", "detector", "CHcpp", "traps<-", "MS.capthist"
-                ), 
-                environment())
+            ), environment())
+            # following are exported only during testing
+            # ,"getD", "untransform", "simpop", "simCH", "ms", "getDetParMat",
+            # "parnames", "getDetDesignData", "covariates", "invlogit",
+            # "usage", "detector", "armaCHcpp", "traps<-", "MS.capthist"
         }
         on.exit(stopCluster(clust))
         clusterSetRNGStream(clust, seed)
     } 
     else {
         clust <- NULL
-        #set.seed(seed)
     }
 
     ####################################################################
@@ -769,12 +812,12 @@ ipsecr.fit <- function (
         colnames(vardesign) <- betanames
         if (ncores > 1) {
             list(...) # evaluate any promises cf boot
-            newsim <- parRapply(clust, vardesign, simfn, 
+            newsim <- parRapply(clust, vardesign, simfn,  
                 distribution = details$distribution, ...)
             newsim <- t(matrix(newsim, ncol = nrow(vardesign)))
         } 
         else {
-            newsim <- t(apply(vardesign,1,simfn, 
+            newsim <- t(apply(vardesign,1,simfn,
                 distribution = details$distribution, ...))
         }
         OK <- apply(!apply(newsim,1, is.na), 2, all)
